//////////////////////////////////////////////////////////////////////////////
//  Author : florent robinet (LAL - Orsay): robinet@lal.in2p3.fr
//////////////////////////////////////////////////////////////////////////////
#ifndef __Segments__
#define __Segments__

#include "CUtils.h"
#include "ReadAscii.h"
#include "TFile.h"
#include "TTree.h"
#include "TH1.h"

using namespace std;

/**
 * Handle time segment lists.
 * Every GW analysis has to deal with segment lists at some point. The class Segments was built to handle segment lists in an easy way. A segment is a GPS interval and the Segments class was designed to handle lists of segments. Once the Segments class is constructed, many functions offer operations on the segments. Convention for segments is described in the <a href="../../Main/convention.html#segments">GWOLLUM convention</a>.
 * \author    Florent Robinet
 */
class Segments {

  friend class ReadTriggerSegments;
  
 public:
  
  /**
   * @name Constructors and destructors
   @{
  */
  /**
   * Constructor of the Segments class.
   * With this constructor, the Segments structure is empty.
   */
  Segments(void);
  
  /**
   * Constructor of the Segments class.
   * A segment list contained in a text file is loaded
   * Note that input segments can overlap in time (overlaps will be internally merged).
   * @param segment_file path to the segment text file. Several formats are supported and are descibed in the <a href="../../Main/convention.html#segments">GWOLLUM convention</a> for segments.
   * @param ncol number of columns in the segment file. It defines the format of the input list. See the <a href="../../Main/convention.html#segments">GWOLLUM convention</a> for supported formats. If ncol<0, the format is automatically detected.
   * @param default_tag a default tag value can be specified to be attributed to every segments. When the segment file has a 5-column format, this default value is not used unless it is >=0. In that case, only segments with tag=default_tag are loaded.
   */
  Segments(const string segment_file, const int ncol=2, const int default_tag=-1);

  /**
   * Constructor of the Segments class.
   * A segment list contained in a TTree structure is loaded.
   * Note that input segments can overlap in time (overlaps will be internally merged).
   * @param segment_tree pointer to the segment TTree structure. See the <a href="../../Main/convention.html#segments">GWOLLUM convention</a> for the TTree container.
   */
  Segments(TTree *segment_tree);

  /**
   * Constructor of the Segments class.
   * A segment list is given as lists of segment starts and ends.
   * The input segments must follow the <a href="../../Main/convention.html#segments">GWOLLUM convention</a> for segments.
   * Note that input segments can overlap in time (overlaps will be internally merged).
   * @param segment_starts list of segment starts (sorted by increasing GPS values).
   * @param segment_ends list of segment ends
   */
  Segments(const vector<double> segment_starts, const vector<double> segment_ends);

  /**
   * Constructor of the Segments class.
   * Constructor for a single segment.
   * The input segment must follow the <a href="../../Main/convention.html#segments">GWOLLUM convention</a> for segments.
   * @param segment_start segment start
   * @param segment_end segment end
   * @param segment_tag segment tag
   */
  Segments(const double segment_start, const double segment_end, const int segment_tag=-1);

  /**
   * Destructor of the Segments class.
   */
  virtual ~Segments(void);  
  /**
     @}
  */


  /**
   * Returns the status of the Segments object.
   * - if false, the object is corrupted and the class functions are likely to mis-behave or even crash.
   * - if true, no problem.
   */ 
  inline bool GetStatus(void) const { return status; };

  /**
   * Returns the number of disjoint time segments in the list.
   */ 
  inline int GetNsegments(void){ return (int)seg_start.size(); };

  /**
   * Returns the number of tagged segments in the list.
   * @param tag tag value of segments to consider.
   */ 
  int GetNSegmentsWithTag(const int tag);

  /**
   * Returns segment starts.
   * This function returns a vector which contains all the segment starts.
   */ 
  inline vector<double> GetStarts(void){ return seg_start; };

  /**
   * Returns segment ends.
   * This function returns a vector which contains all the segment ends.
   */ 
  inline vector<double> GetEnds(void){ return seg_end; };

  /**
   * Returns segment tags.
   * This function returns a vector which contains all the segment tags.
   */ 
  inline vector<int> GetTags(void){ return seg_tag; };

  /**
   * Returns the starting time of segment number s.
   * Returns -1 if not found.
   * @param s segment number
   */ 
  double GetStart(const int s=0);

  /**
   * Returns the ending time of segment number s.
   * Returns -1 if not found.
   * @param s segment number
   */ 
  double GetEnd(const int s=0);

  /**
   * Returns the duration of segment number s.
   * Returns -1 if not found.
   * @param s segment number
   */ 
  double GetDuration(const int s=0);

  /**
   * Returns the tag value of segment number s.
   * Returns -1 if not found.
   * @param s segment number
   */ 
  int GetTag(const int s);

  /**
   * Returns the total livetime value.
   * Every segments in the list are considered.
   */ 
  inline double GetLiveTime(void){ return livetime_tot; };

  /**
   * Returns the livetime value.
   * Only segments between 'gps_start' and 'gps_end' are considered.
   * @param gps_start starting GPS time
   * @param gps_end ending GPS time
   */ 
  double GetLiveTime(const double gps_start, const double gps_end);

  /**
   * Returns the livetime value of segment number s.
   * Returns -1 if not found.
   * @param s segment number
   */ 
  double GetLiveTime(const int s);

  /**
   * Returns the livetime value for segments with a given tag.
   * @param tag tag value
   */ 
  double GetLiveTimeWithTag(const int tag);

  /**
   * Returns an histogram representing the segments.
   * The GPS time is represented on the X axis:
   * - 1 is used when inside a segment.
   * - 0 is used when outside a segment.
   * @param offset time offset applied to segment GPS times.
   * \remark The user is in charge of deleting this histogram.
   * \remark A NULL pointer is returned if there is no segments.
   */ 
  TH1I* GetHisto(const double offset=0.0);

  /**
   * Returns an histogram representing the segments.
   * Only segments between 'gps_start' and 'gps_end' are represented.
   * The GPS time is represented on the X axis:
   * - 1 is used when inside a segment.
   * - 0 is used when outside a segment.
   * @param gps_start starting GPS time of the histogram.
   * @param gps_end ending GPS time of the histogram.
   * @param offset time offset applied to segment GPS times.
   * \remark The user is in charge of deleting this histogram.
   * \remark A NULL pointer is returned if there is no segments.
   */ 
  TH1I* GetHisto(const double gps_start, const double gps_end, const double offset=0.0);

  /**
   * Returns a formatted TTree representing the segments.
   * This tree is formatted as described in the <a href="../../Main/convention.html#segments">GWOLLUM convention</a> for segments.
   * \remark The user is in charge of deleting this TTree.
   * \remark A NULL pointer is returned if failure.
   */ 
  TTree* GetTree(void);

  /**
   * Returns a formatted TTree representing the segments.
   * This tree is formatted as described in the <a href="../../Main/convention.html#segments">GWOLLUM convention</a> for segments.
   * Only segments between 'gps_start' and 'gps_end' are copied in the tree.
   * @param gps_start starting GPS time
   * @param gps_end ending GPS time
   * \remark The user is in charge of deleting this TTree.
   * \remark A NULL pointer is returned if failure.
   */ 
  TTree* GetTree(const double gps_start, const double gps_end);

  /**
   * Returns a time-series representing the segments.
   * The returned vector 'timeseries' is filled with true when inside a segment, false otherwise. The memory pointed by timeseries must be allocated before calling this function. false is returned if this function fails.

   * IMPORTANT: the array pointed by timeseries must be the right size: (gps_end-gps_start)*sampling_rate. No check is performed by this function!
   * @param timeseries pointer to the time-series array 
   * @param sampling_rate sampling rate [Hz]
   * @param gps_start starting GPS time
   * @param gps_end ending GPS time
   */ 
  bool GetTimeSeries(bool *timeseries, const int sampling_rate, const int gps_start, const int gps_end);

  /**
   * Returns the closest segment of a given GPS time.
   * This function fills the vector 'seg' with the segment which is found to be the closest in time of the GPS value gps. As a first priority the segment which contains 'gps' is returned. If 'gps' is not inside a segment, 'seg' is filled with the segment whose lower/upper bound is the closest of 'gps'.
   *
   * In case of success, the segment number is returned.
   * In case of failure, -1 is returned and 'seg' is empty.
   * @param gps GPS time of interest
   * @param seg returned segment as a vector (GPS_START  GPS_END)
   */ 
  int GetClosestSegment(const double gps, vector<double> &seg);

  /**
   * Checks whether a given GPS time is inside a segment of the list.
   * This function returns true if the GPS value gps is found inside a time segment.
   * @param gps GPS time of interest
   */ 
  bool IsInsideSegment(const double gps);

  /**
   * Checks whether a given GPS time is inside a segment of the list.
   * This function returns true if the GPS value gps is found inside a time segment.
   * @param gps GPS time of interest
   * @param seg_n if true is returned, the segment number of the segment containing gps is returned.
   */ 
  bool IsInsideSegment(const double gps, int &seg_n);
      
  /**
   * Apply a padding to segments.
   * This function shifts the segments start and end by a given value:
   * [gps_start; gps_end[ &rarr; [gps_start+pad_low; gps_end+pad_up[
   *
   * The Segments object is automatically re-organized to remove overlaps. Some segments are removed if the duration becomes null or negative.
   *
   * true is returned in case of success.
   * @param pad_low padding added to the segment lower boundary
   * @param pad_up padding added to the segment upper boundary
   */
  bool ApplyPadding(const double pad_low, const double pad_up);

  /**
   * Round segment boundaries to the second.
   * Segments starts are rounded to the previous integer and segments ends are rounded to the next integer.
   *
   * The Segments object is automatically re-organized to remove overlaps.
   *
   * true is returned in case of success.
   */
  bool MakeSecond(void);

  /**
   * Apply a 'smart' time shift.
   * Each segment is time-shited such as the new shifted start is positioned where the segment used to end: 
   * \verbatim 
   BEFORE:       |----|        |-|    |----| |----|
   AFTER:             |----|     |-|       |-|\endverbatim
   * Notice that if the gap is not long enough for the shifted segment, the segment end is adjusted not to overlap the next segment. This way, the resulting segments never overlap the original segments. As a consequence, the resulting livetime is always smaller or equal to the original livetime.
   * The final livetime is returned in case of success, -1.0 if failure.
   */
  double SmartShift(void);

  /**
   * Adds a segment to the list.
   * This function adds the segment [gps_start; gps_end[ to the Segments object. The Segments object is automatically re-organized to remove the overlaps. Optionally the new segment can be tagged.
   *
   * true is returned in case of success.
   * @param gps_start segment start
   * @param gps_end segment end
   * @param tag segment tag
   */
  bool AddSegment(const double gps_start, const double gps_end, const int tag=-1);

  /**
   * Appends segments to the list.
   * This function appends the segments in 'Seg' to the current segment object.
   * This only works if the first segment of 'Seg' starts after the last segment (start) of the current segment object. 
   *
   * true is returned in case of success.
   * @param Seg segment object to append
   */
  bool Append(Segments *Seg);

  /**
   * Intersects two segment lists.
   * This function intersects the current segment list with a second segment list given in argument. This function is destructive and there is no way to go back to the original segment list.
   *
   * true is returned in case of success.
   * @param Seg external Segments objects (not modified)
   */
  bool Intersect(Segments *Seg);

  /**
   * Remove segments after a given time.
   * @param gps GPS time after which to truncate
   */
  bool TruncateAfter(const double gps);

  /**
   * Intersects with a segment.
   * This function intersects the current segment list with a segment given in argument. This function is destructive and there is no way to go back to the original segment list. Optionally, the input segment can be tagged.
   *
   * true is returned in case of success.
   * @param gps_start segment start
   * @param gps_end segment end
   * @param tag segment tag
   */
  bool Intersect(const double gps_start, const double gps_end, const int tag=-1);

  /**
   * Reverses the segment list.
   * This means that the gaps between segments are now considered as segments.
   * \verbatim BEFORE:       |-------|      |-------|  |-|   |----|
   AFTER:    ----|       |------|       |--| |---|    |---- \endverbatim
   * Of course the first and last segment are finite. They start at 700,000,000 and stop at 2,000,000,000 respectively.
   *
   * true is returned in case of success.
   * \remark all tags are lost in the reversing process. The new segments are tagged with -1.
   */
  bool Reverse(void);

  /**
   * Resets the segment list.
   * As a result, the segment list is empty.
   */
  void Reset(void);

  /**
   * Tags segment number s.
   * true is returned in case of success.
   * @param s segment number
   * @param tag new tag value
   */
  bool SetTag(const int s, const int tag);

  /**
   * Tags all segments.
   * true is returned in case of success.
   * @param tag new tag value for all segments
   */
  bool SetTags(const int tag);

  /**
   * Dumps the current segment list in the standard output.
   * Several formats are possible and are given by the 'ncol' argument.
   * See the <a href="../../Main/convention.html#segments">GWOLLUM convention</a> for segments. In addition:
   * - if 'ncol=1' the duration of each segment is printed.
   * - if 'ncol=-1' the tag of each segment is printed.
   *
   * true is returned in case of success.
   * @param ncol number of columns in the segment file. It defines the format of the input list. See the <a href="../../Main/convention.html#segments">GWOLLUM convention</a> for supported formats
   */
  bool Dump(int ncol=2);

  /**
   * Writes segments in a text file.
   * Several formats are supported and are defined by the 'ncol' argument.
   * See the <a href="../../Main/convention.html#segments">GWOLLUM convention</a> for segments.
   *
   * true is returned in case of success.
   * @param txtfilename file name
   * @param ncol number of columns in the segment file. See the <a href="../../Main/convention.html#segments">GWOLLUM convention</a> for supported formats.
   */
  bool Write(const string txtfilename, const int ncol=2);
  
  /**
   * Writes segments in a ROOT file.
   * The segments are saved in a TTree formatted as described in the <a href="../../Main/convention.html#segments">GWOLLUM convention</a> for segments.   
   *
   * true is returned in case of success.
   * @param rootfilename output file name
   */
  bool WriteROOT(const string rootfilename);
    
 protected:
  
  bool status;                ///< status of the Segments object
  vector<double> seg_start;   ///< list of segment starts
  vector<double> seg_end;     ///< list of segment ends
  vector<int> seg_tag;        ///< list of segment tags
  double livetime_tot;        ///< integrated livetime
 
 private:
  
  bool CheckConsistency(void);///< check the segment list consistency 
  bool SetMarks(void);        ///< set time marks for a better speed performance
  bool ComputeLiveTime(void); ///< compute livetime
  bool MergeOverlaps(const int s_start=0);   ///< merge overlapping segments
 
  vector <int> time_mark;     ///< time marks
  vector <int> n_mark;        ///< segment index marks

  inline int  TagProduct(const int tag1, const int tag2){
    if(tag1<0) return -1;
    else if (tag2<0) return -1;
    else return TMath::Max(tag1,tag2);
    return -1; // should never happen
  };                          ///< product of 2 tags

  inline int  TagSum(const int tag1, const int tag2){ return TMath::Max(tag1,tag2); }; ///< sum of 2 tags

  ClassDef(Segments,0)
};

#endif


