//////////////////////////////////////////////////////////////////////////////
//  Author : florent robinet (LAL - Orsay): robinet@lal.in2p3.fr
//////////////////////////////////////////////////////////////////////////////
#ifndef __Streams__
#define __Streams__

#include "TimeDelay.h"
#include "Sample.h"
#include "DetConst.h"

using namespace std;

/**
 * Manage streams.
 * This class was designed to manage data streams as defined in the <a href="../../Main/convention.html#streams">GWOLLUM convention</a> for streams. In particular, the Stream object can be converted into one detector of the LIGO-virgo global network if the stream name is recognized as such. In that case many specific functions are provided to access the detector properties.
 * \author    Florent Robinet
 */
class Streams : public Sample {
  
 public:
  
  /**
   * @name Constructors and destructors
   @{
  */
  /**
   * Constructor of the Streams class.
   * The stream is defined by its name which should respect the stream name convention: [prefix]:[suffix], where [prefix] is 2 characters exactly. If the prefix matches a known LIGO-Virgo detector, the stream is turned into a LV detector. In that case many specific functions are provided to access the detector properties.
   * @param aName stream name
   * @param aVerbose verbosity level
   */
  Streams(const string aName, const int aVerbose=0);
  
  /**
   * Destructor of the ReadCoinc class.
   */
  virtual ~Streams(void);
  /**
     @}
  */

  /**
   * Sets a new name to the stream.   
   * The stream is defined by its name which should respect the stream name convention: [prefix]:[suffix], where [prefix] is 2 characters exactly. If the prefix matches a known LIGO-Virgo detector, the stream is turned into a LV detector.
   * @param aNewName new name
   */
  bool SetName(const string aNewName);

  /**
   * Returns the stream full name.
   */
  inline string GetName(void){ return Name; };

  /**
   * Returns the stream name prefix.
   */
  inline string GetNamePrefix(void){ return NamePrefix; };

  /**
   * Returns the stream name suffix.
   */
  inline string GetNameSuffix(void){ return NameSuffix; };

  /**
   * Returns the stream name suffix with only underscores.
   * Dash characters are turned into underscore characters.
   */
  inline string GetNameSuffixUnderScore(void){ return NameSuffixUnderScore; };

  /**
   * Returns the stream name using the LIGO-Virgo file name convention.
   */
  inline string GetNameConv(void){ return NamePrefix+"-"+NameSuffixUnderScore; };

  /**
   * Returns the stream directory.
   * Convention: ./[name]
   *
   * Optionally, a root directory can be added: [aOutDir]/[name]
   * @param aOutDir root directory
   */
  inline string GetDirectory(const string aOutDir="."){
    return aOutDir+"/"+Name;
  };
    
  /**
   * Returns the detector index for this stream.
   * Returns -1 if this stream was not identified as a LV detector.
   */
  inline int GetDetIndex(void){ return DetIndex; };

  /**
   * Returns the detector prefix for this stream.
   * Returns "" if this stream was not identified as a LV detector.
   */
  inline string GetDetPrefix(void){ if(DetIndex<0) return ""; return DET_PREFIX[DetIndex]; };

  /**
   * Computes F+ and Fx.
   * Returns F+ and Fx values for a source at a specified sky position, polarization angle, and sidereal time. This function only works if the stream has been turned into a LV detector.
   * @param fplus returned F+ value
   * @param fcross returned Fx value
   * @param ra right ascension [rad]
   * @param dec declination [rad]
   * @param psi polarisation angle [rad]
   * @param gmst sideral time [s]
   */
  void GetDetAMResponse(double &fplus, double &fcross, const double ra, const double dec, const double psi, const double gmst);
  
  /**
   * Converts the geocentric time to the detector local time.
   * The light propagation direction must be given by the right ascension and declination. This function only works if the stream has been turned into a LV detector.
   * It returns -1.0 if this function fails.
   * @param aRa right ascension
   * @param aDec declination
   * @param aGeocentricTime geocentric time
   */
  inline double GetLocalTime(const double aRa, const double aDec, const double aGeocentricTime){
    if(DetIndex<0) return -1.0;
    else return aGeocentricTime+TimeDelayFromEarthCenter(DetLoc, aRa, aDec, aGeocentricTime);
  };

 private:
  
  // NAME
  int verbose;                 ///< verbosity level
  string Name;                 ///< Stream name
  string NamePrefix;           ///< Stream prefix
  string NameSuffix;           ///< Stream suffix
  string NameSuffixUnderScore; ///< Stream suffix with only underscores
  
  // LV DETECTOR
  void MakeLVDetector(void);
  int DetIndex;                ///< detector index
  void ComputeDetectorResponse(void);///< spacial response
  double Response[3][3];       ///< detector response
  double DetLoc[3];

  ClassDef(Streams,0)
};

#endif


