//////////////////////////////////////////////////////////////////////////////
//  Author : florent robinet (LAL - Orsay): robinet@lal.in2p3.fr
//////////////////////////////////////////////////////////////////////////////
#include "ffl.h"

ClassImp(ffl)


////////////////////////////////////////////////////////////////////////////////////
ffl::ffl(const string aFrameFileList, const string aPlotStyle, const int aVerbosity){
////////////////////////////////////////////////////////////////////////////////////

  // init
  status_OK=true;
  fInFrameFile=aFrameFileList;
  //fFormat=aFormat;
  fVerbosity=aVerbosity;

  // should be an existing text file
  if(!IsTextFile(fInFrameFile)){
    cerr<<"ffl::ffl: file "<<aFrameFileList<<" cannot be read"<<endl;
    status_OK=false;
  }

  // load file
  if(fVerbosity) cout<<"ffl::ffl: detect file format"<<endl;
  ReadAscii *R = new ReadAscii(fInFrameFile,"s;d;d;s;s");
  if(R->GetNCol()!=5||!R->GetNRow()){
    cerr<<"ffl::ffl: the input frame file list has an unknown format"<<endl;
    status_OK=false;
  }
  if(status_OK){
    double test;
    if(!R->GetElement(test,0,1)||test<700000000) fFormat="lcf";
    else fFormat="ffl";
    if(fVerbosity>1) cout<<"ffl::ffl: detected file format = "<<fFormat<<endl;
  }
  else{
    fFormat="none";
    cerr<<"ffl::ffl: file format "<<fFormat<<" is not supported"<<endl;
    status_OK=false;
  }
  delete R;

  // random id
  srand (time(NULL));
  int randint = rand();
  stringstream tmpstream;
  tmpstream<<randint;
  srandint=tmpstream.str();

  // working frame file
  if(fFormat.compare("ffl")) fWoFrameFile="./ffconvert."+srandint+".ffl"; // converted file
  else fWoFrameFile=fInFrameFile; // original file
  if(fVerbosity>1) cout<<"ffl::ffl: working ffl file = "<<fWoFrameFile<<endl;

  seg = new Segments();

  // plotting
  GP = new GwollumPlot ("ffl", aPlotStyle);
  G = new TGraph();
 
}

////////////////////////////////////////////////////////////////////////////////////
ffl::~ffl(){
////////////////////////////////////////////////////////////////////////////////////
  if(fVerbosity>1) cout<<"ffl::~ffl"<<endl;
  Reset();
  delete seg;
  delete G;
  delete GP;
}


////////////////////////////////////////////////////////////////////////////////////
void ffl::Reset(void){
////////////////////////////////////////////////////////////////////////////////////
  if(fVerbosity) cout<<"ffl::Reset: reset ffl"<<endl;
  ffile.clear();
  channels.clear();
  sampling.clear();
  seg->Reset();
}

////////////////////////////////////////////////////////////////////////////////////
void ffl::SetName(const string aName){
////////////////////////////////////////////////////////////////////////////////////
  string p = GP->GetCurrentStyle();
  delete GP;
  GP = new GwollumPlot (aName, p);
}
  
////////////////////////////////////////////////////////////////////////////////////
double* ffl::GetData(int &aSize, const string channel_name, const double gps_start, const double gps_end, const double val_missing){
////////////////////////////////////////////////////////////////////////////////////
  if(!status_OK){
    cerr<<"ffl::GetData: the ffl object is corrupted"<<endl;
    aSize=0;
    return NULL;
  }
  if(gps_start<700000000){
    cerr<<"ffl::GetData: the requested GPS time does not make sense"<<endl;
    aSize=-1;
    return NULL;
  }
  if(gps_end<=gps_start){
    cerr<<"ffl::GetData: the requested GPS range does not make sense"<<endl;
    aSize=-1;
    return NULL;
  }
  if(seg->GetLiveTime(gps_start,gps_end)!=gps_end-gps_start){
    cerr<<"ffl::GetData: the requested GPS range is not entirely covered by the ffl"<<endl;
    aSize=-2;
    return NULL;
  }
  int chindex;
  int chsampling = GetChannelSampling(channel_name,chindex);
  if(chsampling<=0){
    cerr<<"ffl::GetData: channel "<<channel_name<<" was not found in the data"<<endl;
    aSize=-3;
    return NULL;
  }
  
  // output vector
  double *outvector;
  FrVect *chanvect = NULL; // main vector

  // fill Frame vector
  if(fVerbosity>1) cout<<"ffl::GetData: read data for "<<channel_name<<" between "<<gps_start<<" and "<<gps_end<<endl;
  chanvect = FrFileIGetVectDN(frfile,(char*)(channel_name.c_str()),gps_start,gps_end-gps_start);

  // test output
  if(chanvect==NULL||!chanvect->nData){
    cerr<<"ffl::GetData: data could not be read"<<endl;
    if(chanvect!=NULL) FrVectFree(chanvect);
    aSize=-4;
    return NULL;
  }

  aSize=chanvect->nData;
  if(fVerbosity>1) cout<<"ffl::GetData: "<<aSize<<" samples"<<endl;

  // copy output somewhere else
  outvector = new double [aSize];
  for(int i=0; i<aSize; i++){

    // test for nan (corrupted data)
    if(chanvect->dataD[i]==chanvect->dataD[i]) outvector[i]=chanvect->dataD[i];
    else{
      cerr<<"ffl::GetData: corrupted data"<<endl;
      FrVectFree(chanvect);
      delete outvector;
      aSize=-5;
      return NULL;
    }
  }


  // missing data
  if(chanvect->next!=NULL){

    // return
    if(val_missing==100.0){
      cerr<<"ffl::GetData: missing data"<<endl;
      FrVectFree(chanvect);
      delete outvector;
      aSize=-6;
      return NULL;
    }
 
    // flag missing samples
    for(int i=0; i<aSize; i++){
      if(chanvect->next->dataU[i]==0) outvector[i]=val_missing;
    }
  }
  FrVectFree(chanvect);

  return outvector;
}

////////////////////////////////////////////////////////////////////////////////////
bool ffl::PrintAsciiData(const string channel_name, const double gps_start, const double gps_end, string outputfilename){
////////////////////////////////////////////////////////////////////////////////////
  if(!status_OK){
    cerr<<"ffl::PrintAsciiData: the ffl object is corrupted"<<endl;
    return false;
  }
  
  // get data
  int dvsize;
  double *dv = GetData(dvsize, channel_name, gps_start, gps_end);
  if(dvsize<=0) return false;

  // fill outputfile
  ofstream outputfile(outputfilename.c_str());
  if(outputfile.fail()){
    cerr<<"ffl::PrintAsciiData cannot open output file" <<endl;
    return false;
  }

  for(int i=0; i<dvsize; i++) 
    outputfile<<setprecision(16)<<gps_start+((double)i+0.5)*(gps_end-gps_start)/(double)dvsize<<" "<<setprecision(6)<<dv[i]<<endl;
  delete dv;

  outputfile.close();

  return true;
}

////////////////////////////////////////////////////////////////////////////////////
TGraph* ffl::PlotData(const string channel_name, const double gps_start, const double gps_end){
////////////////////////////////////////////////////////////////////////////////////
  if(!status_OK){
    cerr<<"ffl::PlotData: the ffl object is corrupted"<<endl;
    return NULL;
  }
  
  // get data
  int dvsize;
  double *dv = GetData(dvsize, channel_name, gps_start, gps_end);
  if(dvsize<=0) return NULL;

  // fill graph
  delete G;
  G = new TGraph(dvsize);
  for(int i=0; i<dvsize; i++) G->SetPoint(i,gps_start+((double)i+0.5)*(gps_end-gps_start)/(double)dvsize,dv[i]);
  delete dv;

  // plot
  G->GetHistogram()->GetXaxis()->SetNoExponent();
  G->GetHistogram()->GetXaxis()->SetNdivisions(5,5,0);
  G->GetHistogram()->GetXaxis()->SetTitle("Time [s]");
  G->GetHistogram()->SetTitle((channel_name+" time series").c_str());
  G->GetHistogram()->GetYaxis()->SetTitle("Amplitude [?]");
  GP->Draw(G,"APL");
  GP->SetGridx();  GP->SetGridy();

  return G;
}

////////////////////////////////////////////////////////////////////////////////////
TGraph* ffl::PlotData2(const string channel_name1, const string channel_name2, const double gps_start, const double gps_end){
////////////////////////////////////////////////////////////////////////////////////
  if(!status_OK){
    cerr<<"ffl::PlotData2: the ffl object is corrupted"<<endl;
    return NULL;
  }
  
  // get data
  int dvsize1, dvsize2;
  double *dv1 = GetData(dvsize1, channel_name1, gps_start, gps_end);
  if(dvsize1<=0) return NULL;
  double *dv2 = GetData(dvsize2, channel_name2, gps_start, gps_end);
  if(dvsize2<=0){ delete dv1; return NULL; }

  if(dvsize1!=dvsize2){
    delete dv1; delete dv2;
    return NULL;
  }

  // fill graph
  delete G;
  G = new TGraph(dvsize1);
  for(int i=0; i<dvsize1; i++) G->SetPoint(i,gps_start+((double)i+0.5)*(gps_end-gps_start)/(double)dvsize1,sqrt(dv1[i]*dv1[i]+dv2[i]*dv2[i]));
  delete dv1;
  delete dv2;

  // plot
  G->GetHistogram()->GetXaxis()->SetNoExponent();
  G->GetHistogram()->GetXaxis()->SetNdivisions(5,5,0);
  G->GetHistogram()->GetXaxis()->SetTitle("Time [s]");
  G->GetHistogram()->SetTitle(("#sqrt{"+channel_name1+"^2+"+channel_name2+"^2} time series").c_str());
  G->GetHistogram()->GetYaxis()->SetTitle("Amplitude [?]");
  GP->Draw(G,"APL");
  GP->SetGridx();  GP->SetGridy();

  return G;
}

////////////////////////////////////////////////////////////////////////////////////
TGraph* ffl::PlotDerivative(const string channel_name, const double gps_start, const double gps_end){
////////////////////////////////////////////////////////////////////////////////////
  if(!status_OK){
    cerr<<"ffl::PlotDerivative: the ffl object is corrupted"<<endl;
    return NULL;
  }
  
  // get data
  int dvsize;
  double *dv = GetData(dvsize, channel_name, gps_start, gps_end);
  if(dvsize<=0) return NULL;

  // fill graph
  delete G;
  G = new TGraph(dvsize-1);
  for(int i=0; i<dvsize-1; i++) G->SetPoint(i,
					    gps_start+(2.0*i+1.0)/2.0*(gps_end-gps_start)/(double)dvsize,
					    (dv[i+1]-dv[i])/(gps_end-gps_start)*(double)dvsize);
  delete dv;

  // plot
  G->GetHistogram()->GetXaxis()->SetNoExponent();
  G->GetHistogram()->GetXaxis()->SetNdivisions(5,5,0);
  G->GetHistogram()->GetXaxis()->SetTitle("Time [s]");
  G->GetHistogram()->SetTitle((channel_name+" time derivative").c_str());
  G->GetHistogram()->GetYaxis()->SetTitle("Derivative amplitude [?]");
  GP->Draw(G,"APL");
  GP->SetGridx();  GP->SetGridy();

  return G;
}

////////////////////////////////////////////////////////////////////////////////////
bool ffl::LoadFrameFile(const int gps){
////////////////////////////////////////////////////////////////////////////////////
  if(!status_OK){
    cerr<<"ffl::Loadffl: The ffl object is corrupted"<<endl;
    return false;
  }

  // if LCF -> convert first
  if(fFormat.compare("ffl")&&!Convert2Ffl().compare("none")){
    cerr<<"ffl::Loadffl: the conversion to FFL failed"<<endl;
    return false;
  }

  // reset
  Reset();

  // define expected column format for FFL convention
  string form="s;d;d;s;s";

  // load file
  if(fVerbosity) cout<<"ffl::LoadFrameFile: load FFL file"<<endl;
  ReadAscii *R = new ReadAscii(fWoFrameFile,form);
  if(R->GetNCol()!=5){
    cerr<<"ffl::Loadffl: the frame file has an unknown format"<<endl;
    delete R;
    return false;
  }

  // extract information
  vector<double> start;  // list of start
  vector<double> stop;   // list of stop
  vector<double> dur;   // list of durations
  if(fVerbosity) cout<<"ffl::LoadFrameFile: extract frame information"<<endl;
  if(!R->GetCol(ffile,0)){
    cerr<<"ffl::Loadffl: cannot extract list of files"<<endl;
    delete R;
    return false;
  }
  
  if(!R->GetCol(start,1)){
    cerr<<"ffl::Loadffl: cannot extract list of starts"<<endl;
    delete R;
    return false;
  }
  if(!R->GetCol(dur,2)){
    cerr<<"ffl::Loadffl: cannot extract list of durations"<<endl;
    delete R;
    return false;
  }
  for(int l=0; l<(int)start.size(); l++) stop.push_back(start[l]+dur[l]);
  dur.clear();

  if(fVerbosity) cout<<"ffl::LoadFrameFile: check frame information"<<endl;
    
  // make segments
  delete seg;
  seg = new Segments(start,stop);
  start.clear();
  stop.clear();
  if(!seg->GetStatus()) return false;

  // make Fr file
  frfile = FrFileINew((char*)fWoFrameFile.c_str());

  // get list of channels (first frame file!)
  if(!ExtractChannels(gps)){
    cerr<<"ffl::Loadffl: cannot extract channel information"<<endl;
    return false;
  }

  return seg->GetStatus();
}

////////////////////////////////////////////////////////////////////////////////////
int ffl::GetChannelSampling(const string channel_name, int &channel_index){
////////////////////////////////////////////////////////////////////////////////////
  if(!status_OK){
    cerr<<"ffl::GetChannelSampling: The ffl object is corrupted"<<endl;
    return -1;
  }
  
  for(channel_index=0; channel_index<(int)channels.size(); channel_index++){
    if(!channel_name.compare(channels[channel_index])) return sampling[channel_index];
  }

  return 0;
}

////////////////////////////////////////////////////////////////////////////////////
vector <string> ffl::GetChannelList(const string channel_filter){
////////////////////////////////////////////////////////////////////////////////////
  if(!channel_filter.compare("")) return channels;

  vector <string> subchannels;

  // get sub-strings
  vector <string> substring = SplitString(channel_filter, ' ');

  // loop over channels
  for(int c=0; c<(int)channels.size(); c++){

    // check pattern
    for(int s=0; s<(int)substring.size(); s++){
      if(!fnmatch(substring[s].c_str(), channels[c].c_str(),FNM_NOESCAPE)){
	subchannels.push_back(channels[c]);
	break;
      }
    }
  }

  return subchannels;
}

////////////////////////////////////////////////////////////////////////////////////
bool ffl::IsChannel(const string channel_name){
////////////////////////////////////////////////////////////////////////////////////
  for(int l=0; l<(int)channels.size(); l++){
    if(!channel_name.compare(channels[l])) return true;
  }
  return false;
}

////////////////////////////////////////////////////////////////////////////////////
string ffl::Convert2Ffl(void){
////////////////////////////////////////////////////////////////////////////////////
  if(!status_OK){
    cerr<<"ffl::Convert2Ffl: The ffl object is corrupted"<<endl;
    return "none";
  }
  if(!fFormat.compare("ffl")){
    cerr<<"ffl::Convert2Ffl: the frame file is already defined with a FFL format"<<endl;
    return "none";
  }

  // load file
  if(fVerbosity) cout<<"ffl::Convert2Ffl: FFL conversion"<<endl;
  ReadAscii *R = new ReadAscii(fInFrameFile,"s;s;d;d;s");
  if(R->GetNCol()!=5){
    cerr<<"ffl::Convert2Ffl: input file "<<fInFrameFile<<" does not look like a LCF file"<<endl;
    delete R;
    return "none";
  }

  vector<double> start_tmp;
  vector<double> dur_tmp;
  vector<string> ffile_tmp;

  // extract data
  if(!R->GetCol(ffile_tmp,4)){
    cerr<<"ffl::Loadffl: cannot extract list of files"<<endl;
    delete R;
    return "none";
  }
  if(!R->GetCol(start_tmp,2)){
    cerr<<"ffl::Convert2Ffl: cannot extract list of starts"<<endl;
    delete R;
    return "none";
  }
  if(!R->GetCol(dur_tmp,3)){
    cerr<<"ffl::Convert2Ffl: cannot extract list of durations"<<endl;
    delete R;
    return "none";
  }

  // fill converted file
  string fff1, fff2;
  ofstream outfile(fWoFrameFile.c_str());
  for(int l=0; l<(int)start_tmp.size(); l++){
    fff1.clear();
    fff2.clear();
    if(!(ffile_tmp[l].substr(0,5)).compare("file:")) fff1=ffile_tmp[l].substr(5);// remove "file:"
    else fff1=ffile_tmp[l];
    if(!(fff1.substr(0,11)).compare("//localhost")) fff2=fff1.substr(11);// remove "//localhost"
    else fff2=fff1;
    outfile<<fff2<<" "<<setprecision(15)<<start_tmp[l]<<" "<<dur_tmp[l]<<" 0 0"<<endl;
  }
  outfile.close();
  
  delete R;
  start_tmp.clear();
  dur_tmp.clear();
  ffile_tmp.clear();

  return fWoFrameFile;
}

////////////////////////////////////////////////////////////////////////////////////
bool ffl::DefineTmpDir(const string aTmpDirPath){
////////////////////////////////////////////////////////////////////////////////////
  if(!IsDirectory(aTmpDirPath)){
    cerr<<"ffl::DefineTmpDir: tmp directory is missing"<<endl;
    return false;
  }
  if(fFormat.compare("ffl")) fWoFrameFile=aTmpDirPath+"/ffconvert."+srandint+".ffl";
  else fWoFrameFile=fInFrameFile;
  return true;
}

////////////////////////////////////////////////////////////////////////////////////
bool ffl::ExtractChannels(const int gps){
////////////////////////////////////////////////////////////////////////////////////
  if(!status_OK){
    cerr<<"ffl::ExtractChannels: The ffl object is corrupted"<<endl;
    return false;
  }
  if(gps>0&&!seg->IsInsideSegment(gps)){
    cerr<<"ffl::ExtractChannels: the requested GPS time is not covered by the ffl"<<endl;
    return false;
  }
  char* clist = NULL;
  if(gps<=0) clist = FrFileIGetChannelList(frfile, 0);
  else clist = FrFileIGetChannelList(frfile, gps);
  if(clist==NULL){
    cerr<<"ffl::ExtractChannels: no channel returned"<<endl;
    return false;
  }
  string slist = (string)clist;
  delete clist;
  channels.clear();
  sampling.clear();

  // split into lines
  vector <string> lines = SplitString(slist, '\n');
  
  // read each line
  vector <string> words;
  double tmp;
  for(int l=0; l<(int)lines.size(); l++){
    words=SplitString(lines[l], ' ');
    if(words.size()<3){
      words.clear();
      continue;
    }

    // search for ADC
    if(!words[0].compare("ADC")){
      tmp=atof(words[2].c_str());
      if(tmp&&tmp==floor(tmp)){
	channels.push_back(words[1]);
	sampling.push_back((int)tmp);
      }
    }

    // search for PROC
    if(!words[0].compare("PROC")){
      tmp=atof(words[2].c_str());
      if(tmp&&tmp==floor(tmp)){
	channels.push_back(words[1]);
	sampling.push_back((int)tmp);
      }
    }

    // search for SIM
    if(!words[0].compare("SIM")){
      tmp=atof(words[2].c_str());
      if(tmp&&tmp==floor(tmp)){
	channels.push_back(words[1]);
	sampling.push_back((int)tmp);
      }
    }

    // search for SER
    if(!words[0].compare("SER")){
      tmp=atof(words[2].c_str());
      if(tmp&&tmp==floor(tmp)){
	for(int w=3; w<(int)words.size(); w+=2){
	  channels.push_back(words[1]+"_"+words[w]);
	  sampling.push_back((int)tmp);
	}
      }
    }
    
    words.clear();
  }

  if(fVerbosity>2){
    cout<<"Extracted information = "<<endl;
    for(int l=0; l<(int)channels.size(); l++) cout<<channels[l]<<" "<<sampling[l]<<endl;
  }
  return true;
}

