//////////////////////////////////////////////////////////////////////////////
//  Author : florent robinet (LAL - Orsay): robinet@lal.in2p3.fr
//////////////////////////////////////////////////////////////////////////////
#ifndef __GwollumPlot__
#define __GwollumPlot__

#include "CUtils.h"
#include "TStyle.h"
#include "TLine.h"
#include "TMultiGraph.h"
#include "TGraph.h"
#include "TGraph2D.h"
#include "TF1.h"
#include "TH1.h"
#include "TH2.h"
#include "TPad.h"
#include "TCanvas.h"
#include "TLegend.h"
#include "TText.h"

using namespace std;

/**
 * Interface to draw and print ROOT objects.
 * The GwollumPlot class provides a single interface to draw and print ROOT objects like histograms, graphs and so on. It also includes different pre-defined plotting styles like the <i>GWOLLUM style</i>. This class mainly includes wrappers for ROOT drawing functions.
 * 
 * Several pre-defined styles are available. At any moment, you can switch from one style to the next.
 * <ul>
 * <li><b>&quot;GWOLLUM&quot;</b>: This is the full GWOLLUM style. Backgrounds are black and deep-blue sea markings are used.</li>
 * <li><b>&quot;PINK&quot;</b>: Backgrounds are black and pink markings are used.</li>
 * <li><b>&quot;FIRE&quot;</b>: This style is convenient for black and white printing.</li>
 * <li><b>&quot;STANDARD&quot;</b>: This is the standard ROOT style.</li>
 * </ul>
 * \author    Florent Robinet
 */
class GwollumPlot{

 public:
  
  /**
   * @name Constructors and destructors
   @{
  */

  /**
   * Constructor of the GwollumPlot class.
   * A unique name must be provided with 'aName'. You can also choose a style between:
   * <ul>
   * <li>&quot;GWOLLUM&quot; (default)</li>
   * <li>&quot;WHITEGWOLLUM&quot;</li>
   * <li>&quot;STANDARD&quot;</li>
   * </ul>
   * This style can be changed later.
   * @param aName name to identify the canvas
   * @param aStyleName style name
   */
  GwollumPlot(const string aName, const string aStyleName="GWOLLUM");

  /**
   * Destructor of the GwollumPlot class.
   */
  virtual ~GwollumPlot(void);

  /**
     @}
  */


  /**
   * Adds text in the plot.
   * The position of the plot should be given in the pad coordinates (0-1).
   * @param aText text to add
   * @param aX X coordinate where to start the text
   * @param aY Y coordinate where to start the text
   * @param aSize text size
   * @param aPadIndex pad index when using subpads (see DivideCanvas())
   */
  void AddText(const string aText, const double aX, const double aY, const double aSize, const int aPadIndex=0);

  /**
   * Updates text in the plot.
   * See AddText().
   * @param aText updated text
   */
  void UpdateText(const string aText);

  /**
   * Adds a legend entry.
   * @param aObj pointer to ROOT object to which the legend entry applies
   * @param aLabel label of the legend entry
   * @param aStyle legend style "L" for line, "P" for point and "F" for fill
   */
  void AddLegendEntry(const TObject *aObj, const string aLabel, const string aStyle="LPF");
  
  /**
   * Adds a legend header.
   * @param aLabel label of the legend header
   */
  void AddLegendHeader(const string aLabel);

  /**
   * Draws current legend box.
   */
  void DrawLegend(void);

  /**
   * Resets and removes current legend box.
   */
  void ResetLegend(void){ Wleg->Clear(); Wpad->SetPad(0,0,1,1); Wpad->Modified(1); Wpad->Update();};
  
  /**
   * Draws ROOT object.
   * @param aObj pointer to ROOT object to be drawn
   * @param aOptions drawing options (ROOT options)
   * @param aPadIndex pad index when using subpads (see DivideCanvas())
   */
  inline void Draw(TObject *aObj, const string aOptions="", const int aPadIndex=0){ Wpad->cd(aPadIndex); aObj->Draw(aOptions.c_str()); gPad->Modified(); gPad->Update();};   

  /**
   * UnDraws a ROOT object.
   * @param aObj pointer to ROOT object to be remove
   * @param aPadIndex pad index when using subpads (see DivideCanvas())
   */
  inline void UnDraw(TObject *aObj, const int aPadIndex=0){ Wpad->cd(aPadIndex); gPad->GetListOfPrimitives()->Remove(aObj); gPad->Modified(); gPad->Update();};
  
  /**
   * Prints current canvas in a file.
   * @param aFileName file name to save the current canvas
   */
  inline void Print(const string aFileName){ Wcan->cd(); Wcan->Print(aFileName.c_str()); }; 

  /**
   * Prints current canvas in a file with a rescaling factor.
   * Typically this function is useful when one wants to produce thumbnails.
   * The size of the working canvas is not changed, only the size of the printed object is modified.
   * The width and height are rescaled by 'rescale'.
   * @param aFileName file name to save the current canvas
   * @param aScaleFactor scaling factor
   */
  void Print(const string aFileName, const double aScaleFactor);
  
  /**
   * Prints current canvas in a file with a new size.
   * The size of the working canvas is not changed, only the size of the printed object is modified.
   * @param aFileName file name to save the current canvas
   * @param aNewWidth new canvas horizontal size.
   * @param aNewHeight new canvas vertical size.
   */
  void Print(const string aFileName, const int aNewWidth, const int aNewHeight);

  /**
   * Set Lin/Log scale for X.
   * - aValue = 0 X scale will be linear.
   * - aValue = 1 X scale will be logarithmic (base 10).
   * - aValue > 1 reserved for possible support of base e or otherSet logarithmic scale for x axis.
   * @param aValue
   * @param aPadIndex pad index when using subpads (see DivideCanvas())
   */
  inline void SetLogx(const int aValue, const int aPadIndex=0){ Wpad->cd(aPadIndex); gPad->SetLogx(aValue); gPad->Modified(); gPad->Update(); }; 

  /**
   * Set Lin/Log scale for Y.
   * - aValue = 0 Y scale will be linear.
   * - aValue = 1 Y scale will be logarithmic (base 10).
   * - aValue > 1 reserved for possible support of base e or otherSet logarithmic scale for y axis.
   * @param aValue
   * @param aPadIndex pad index when using subpads (see DivideCanvas())
   */
  inline void SetLogy(const int aValue, const int aPadIndex=0){ Wpad->cd(aPadIndex); gPad->SetLogy(aValue); gPad->Modified(); gPad->Update(); }; 

  /**
   * Set Lin/Log scale for Z.
   * - aValue = 0 Z scale will be linear.
   * - aValue = 1 Z scale will be logarithmic (base 10).
   * - aValue > 1 reserved for possible support of base e or otherSet logarithmic scale for z axis.
   * @param aValue
   * @param aPadIndex pad index when using subpads (see DivideCanvas())
   */
  inline void SetLogz(const int aValue, const int aPadIndex=0){ Wpad->cd(aPadIndex); gPad->SetLogz(aValue); gPad->Modified(); gPad->Update(); }; 

  /**
   * Set/Unset grid for X.
   * @param aValue
   * @param aPadIndex pad index when using subpads (see DivideCanvas())
   */
  inline void SetGridx(const int aValue=1, const int aPadIndex=0){ Wpad->cd(aPadIndex); gPad->SetGridx(aValue);  gPad->Modified(); gPad->Update(); }; 

  /**
   * Set/Unset grid for Y.
   * @param aValue
   * @param aPadIndex pad index when using subpads (see DivideCanvas())
   */
  inline void SetGridy(const int aValue=1, const int aPadIndex=0){ Wpad->cd(aPadIndex); gPad->SetGridy(aValue);  gPad->Modified(); gPad->Update(); }; 

  /**
   * Redraw the frame axis.
   * @param aPadIndex pad index when using subpads (see DivideCanvas())
   */
  inline void RedrawAxis(Option_t* option="", int aPadIndex=0){ Wpad->cd(aPadIndex); gPad->RedrawAxis(option); gPad->Modified(); gPad->Update(); }; 

  /**
   * Resizes the global canvas to new dimensions.
   * Set Width and Height of canvas to aWidth and aHeight respectively.
   * If aWidth and/or aHeight are greater than the current canvas window a scroll bar is automatically generated.
   * Use this function to zoom in a canvas and navigate via the scroll bars.
   * @param aWidth new canvas width
   * @param aHeight new canvas height
   */
  inline void ResizePlot(const int aWidth, const int aHeight){
    Wcan->SetWindowSize(aWidth + (aWidth- Wcan->GetWw()), aHeight + (aHeight - Wcan->GetWh()));
    Wcan->SetCanvasSize(aWidth,aHeight);
    Wcan->Modified(); Wcan->Update();
  }; 
  
  /**
   * Returns the canvas width.
   */
  inline int GetWidth(void){ return Wcan->GetWw(); }; 
  
  /**
   * Returns the canvas height.
   */
  inline int GetHeight(void){ return Wcan->GetWh(); }; 
  
  /**
   * Divides current canvas into sub-pads.
   * @param aNpads number of pads
   */
  inline void DivideCanvas(const int aNpads){ Wpad->DivideSquare(aNpads); };

  /**
   * Divides current canvas into sub-pads.
   * @param aNx number of columns
   * @param aNy number of rows
   */
  inline void DivideCanvas(const int aNx, const int aNy){ Wpad->Divide(aNx,aNy); };

  /**
   * Returns current style name
   */
  inline string GetCurrentStyle(void){ return stylename; };
    
  /**
   * Return color number i in current palette.
   */
  inline int GetColorPalette(const int i){ return Wstyle->GetColorPalette(i); };
    
  /**
   * Return the number of colors in current palette.
   */
  inline int GetNumberOfColors(void){ return Wstyle->GetNumberOfColors(); };
    

 protected:
  
  int randid;                      ///< random integer id
  string srandid;                  ///< random string id

  // STYLE
  TStyle *Wstyle;                  ///< working style
  string stylename;                ///< style name

  // CANVAS/PAD
  TCanvas *Wcan;                   ///< working canvas
  TPad    *Wpad;                   ///< working pad

  // ELEMENTS
  TLegend *Wleg;                   ///< plot legends
  TText   *Wtext;                  ///< add text

 private:

  void SetGwollumStyle(void);      ///< sets GWOLLUM style
  void SetPinkStyle(void);         ///< sets PINK style
  void SetFireStyle(void);         ///< sets FIRE style
  void SetStandardStyle(void);     ///< sets STANDARD style
  


  ClassDef(GwollumPlot,0)
};

#endif


